// Copyright (c) 2016, Dart Gaming authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:math';

import 'package:meta/meta.dart';
import 'package:quiver/core.dart';

/// A 360 degree analog stick on a gamepad.
///
/// To listen to changes, simply listen as a stream:
/// ```dart
/// void useAxes(Axes axes) {
///   axes.listen((state) => ...);
/// }
/// ```
///
/// Axes are _not_ guaranteed to be stateful; clients implementing a game loop
/// will want to cache the last received value(s) before they do an update.
///
/// TODO: Provide a helper for game loops that does exactly that.
///
/// There are additional _helper_ streams such as [angleDirection] that only
/// emit when a direction change occurs; this is provided only to help simplify
/// controls for basic games.
///
/// Clients may extend, implement, or mixin this class freely.
abstract class Axes implements Stream<AxesState> {
  /// A stream that only emits when the angled direction changes.
  Stream<AxesDirection> get angleDirection {
    return map((state) => AxesDirection.forAngle(state.angle)).distinct();
  }

  /// A stream that only emits when the x-axis direction changes.
  ///
  /// Valid events are [AxesDirection.left] and [AxesDirection.right].
  Stream<AxesDirection> get xAxisDirection {
    return map((state) => AxesDirection.forXAxis(state.angle)).distinct();
  }

  /// A stream that only emits when the y-axis direction changes.
  ///
  /// Valid events are [AxesDirection.up] and [AxesDirection.down].
  Stream<AxesDirection> get yAxisDirection {
    return map((state) => AxesDirection.forYAxis(state.angle)).distinct();
  }
}

/// A computed direction of an [AxesState].
///
/// While not every game uses absolute directions, this is provided for a simple
/// view of angled and x/y-axis based directions for games that do not use more
/// information.
///
/// Clients should _not_ extend, implement, or mixin this class.
class AxesDirection {
  /// Converts degrees to radians.
  static num _toRadians(num degrees) => degrees * (180 / PI);

  static const up = const AxesDirection._('up');
  static const right = const AxesDirection._('right');
  static const down = const AxesDirection._('down');
  static const left = const AxesDirection._('left');

  // Angular direction
  //     \  UP /
  //      \   /
  // LEFT       RIGHT
  //      /   \
  //     /DOWN \
  //

  /// Returns the angled direction given an angle in [degrees].
  static AxesDirection forAngle(num degrees) {
    if (degrees >= 180 + 45) {
      return AxesDirection.left;
    }
    if (degrees > 180 - 45) {
      return AxesDirection.down;
    }
    if (degrees >= 45) {
      return AxesDirection.right;
    }
    return AxesDirection.up;
  }

  // Plain directions
  //    UP                 |
  // _______               | RIGHT
  //                  LEFT |
  //   DOWN                |

  /// Returns the x-axis direction (left or right) for a given angle.
  static AxesDirection forXAxis(num degrees) {
    return degrees > 180 ? AxesDirection.left : AxesDirection.right;
  }

  /// Returns the y-axis direction (up or down) for a given angle.
  static AxesDirection forYAxis(num degrees) {
    return degrees <= 90 || degrees >= 270
        ? AxesDirection.up
        : AxesDirection.down;
  }

  /// Human readable name.
  final String name;

  const AxesDirection._(this.name);

  @override
  String toString() => name;
}

/// Represents the state of an [Axes] at a point in time.
class AxesState {
  /// An angle of `0` in degrees (often interpreted as full "up").
  static const angleUp = 0;

  /// An angle of `90` in degrees (often interpreted as full "right").
  static const angleRight = 90;

  /// An angle of `180` in degrees (often interpreted as full "down").
  static const angleDown = 180;

  /// An angle of `270` in degrees (often interpreted as full "left").
  static const angleLeft = 270;

  /// A suggested [force] when the stick has no direction set.
  static const forceWhenUnused = 0;

  /// A suggested [force] when the stick is in use but force is not measurable.
  static const forceWhenInUse = 1;

  /// A suggested [pressure] when not measurable.
  static const pressureDefault = 1;

  /// Angle of the stick, in degrees.
  final num angle;

  /// A strength, in percent, of the directional force.
  ///
  /// May simply always be `1` or `0` if not measurable.
  final num force;

  /// Represents when this state was created.
  ///
  /// May be `null` if not supported by the underlying platform.
  final num timestamp;

  /// Pressure applied downwards.
  ///
  /// May simply always be `1` if not measurable.
  final num pressure;

  /// Create a new axes state.
  @literal
  const AxesState({
    @required this.angle,
    this.force: forceWhenInUse,
    this.timestamp,
    this.pressure: pressureDefault,
  });

  /// Create a new axes state to signify the joystick is not in use.
  @literal
  const AxesState.notInUse({this.timestamp})
      : angle = angleUp,
        force = forceWhenUnused,
        pressure = pressureDefault;

  /// Create a new axes state to signify a simple "up" direction.
  @literal
  const AxesState.up({
    this.force: forceWhenInUse,
    this.timestamp,
    this.pressure: pressureDefault,
  })
      : angle = angleUp;

  /// Create a new axes state to signify a simple "right" direction.
  @literal
  const AxesState.right({
    this.force: forceWhenInUse,
    this.timestamp,
    this.pressure: pressureDefault,
  })
      : angle = angleRight;

  /// Create a new axes state to signify a simple "down" direction.
  @literal
  const AxesState.down({
    this.force: forceWhenInUse,
    this.timestamp,
    this.pressure: pressureDefault,
  })
      : angle = angleDown;

  /// Create a new axes state to signify a simple "left" direction.
  @literal
  const AxesState.left({
    this.force: forceWhenInUse,
    this.timestamp,
    this.pressure: pressureDefault,
  })
      : angle = angleLeft;

  @override
  bool operator ==(Object o) {
    if (o is AxesState) {
      return angle == angle &&
          force == force &&
          pressure == pressure &&
          timestamp == timestamp;
    }
    return false;
  }

  @override
  int get hashCode => hash2(hash2(angle, force), hash2(pressure, timestamp));

  @override
  String toString() => ''
      '$AxesState {\n'
      '  angle:     $angle\n'
      '  force:     $force\n'
      '  timestamp: $timestamp\n'
      '  pressure:  $pressure\n'
      '}';
}
