// Copyright (c) 2016, Dart Gaming authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import 'package:gamepad/src/drivers/api/axes.dart';

/// A virtual axes receives all events from itself (via [StreamSink] interface).
///
/// This is the recommending underlying implementation for an [Axes], due to the
/// fact that it already implements the entire interface. New implementations
/// should "hide" that they are a virtual axes:
/// ```dart
/// abstract class KeyboardAxes implements Axes {
///   factory KeyboardAxes() = _VirtualKeyboardAxes;
/// }
///
/// class _VirtualKeyboardAxes extends VirtualAxes implements KeyboardAxes {}
/// ```
abstract class VirtualAxes<E> extends Stream<AxesState>
    with Axes
    implements StreamSink<AxesState> {
  final StreamController<AxesState> _controller;

  VirtualAxes() : _controller = new StreamController<AxesState>();

  /// Activates the axes, given an optional [context].
  ///
  /// A context may contain information such as where in the UI to create the
  /// axes, a facade to control another axes, or even a channel to activate a
  /// remote connection and receive commands.
  Future<Null> activate([E context]) => new Future<Null>.value();

  /// Deactivates the axes.
  ///
  /// A deactivated axes should close its stream on any existing listeners.
  Future<Null> deactivate() => new Future<Null>.value();

  @override
  void add(AxesState event) => _controller.add(event);

  @override
  void addError(error, [StackTrace stackTrace]) {
    _controller.addError(error, stackTrace);
  }

  @override
  Future addStream(Stream<AxesState> stream) => _controller.addStream(stream);

  @override
  Future close() => _controller.close();

  @override
  Future get done => _controller.done;

  @override
  StreamSubscription<AxesState> listen(
    void onData(AxesState event), {
    Function onError,
    void onDone(),
    bool cancelOnError,
  }) {
    return _controller.stream.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }
}

/// A mixin for [StreamSink<AxesState>] implementation that has simple triggers.
///
/// ## Example use
/// ```dart
/// class SimpleVirtualAxes = VirtualAxes with TriggerDirection;
/// ```
abstract class TriggerDirection implements StreamSink<AxesState> {
  /// Fires an "up" state.
  void triggerUp() => add(const AxesState.up());

  /// Fires a "right" state.
  void triggerRight() => add(const AxesState.right());

  /// Fires a "down" state.
  void triggerDown() => add(const AxesState.down());

  /// Fires a "left" state.
  void triggerLeft() => add(const AxesState.left());
}
